/*
 * DISCLAIMER:
 *   WITH RESPECT TO THE INFORMATION, CODE AND/OR EXECUTABLES PROVIDED  HEREIN,
 *   BRASHEAR LP MAKES NO WARRANTIES OF ANY KIND, EITHER EXPRESSED OR IMPLIED,
 *   INCLUDING BUT NOT LIMITED TO THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *   FITNESS FOR A PARTICULAR PURPOSE. IN ADDITION, IN NO EVENT SHALL BRASHEAR LP
 *   BE HELD LIABLE FOR ANY DAMAGES WHATSOEVER INCLUDING DIRECT, INDIRECT, INCIDENTAL,
 *   CONSEQUENTIAL, LOSS OF BUSINESS PROFITS OR SPECIAL DAMAGES, EVEN IF
 *   BRASHEAR LP HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.
*/




#include <stdio.h>
#include <stdlib.h>
#include <stddef.h>
#include <unistd.h>
#include <time.h>
#include <i86.h>
#include <sys/pci.h>
#include <fcntl.h>
#include <sys/mman.h>
#include <netinet/in.h>
#include <sys/proxy.h>
#include <sys/irqinfo.h>

#include <process_proto.h>

#define DUTY_CYCLE	2

/* PCI data */
#define DATUM_PCI_VENDOR_ID	0x12E2
#define DATUM_PCI_DEVICE_ID	0x4013
#define DATUM_DPRAM_TOP		0x800

/* Device register control field bit definitions */
#define DATUM_CONTROL_BIT_LOCKEN	0x01
#define DATUM_CONTROL_BIT_EVSOURCE	0x02
#define DATUM_CONTROL_BIT_EVSENSE	0x04
#define DATUM_CONTROL_BIT_EVENTEN	0x08
#define DATUM_CONTROL_BIT_STREN		0x10
#define DATUM_CONTROL_BIT_STRMODE	0x20
#define DATUM_CONTROL_BIT_FREQSEL0	0x40
#define DATUM_CONTROL_BIT_FREQSEL1	0x80

/* Device register interrupt mask and status field bit definitions */
#define DATUM_INT_BIT_EVENT		0x01
#define DATUM_INT_BIT_PERIOD	0x02
#define DATUM_INT_BIT_STROBE	0x04
#define DATUM_INT_BIT_1PPS		0x08
#define DATUM_INT_BIT_GPS		0x10
#define DATUM_INT_BIT_ALL		( DATUM_INT_BIT_EVENT |\
								  DATUM_INT_BIT_PERIOD|\
								  DATUM_INT_BIT_STROBE|\
								  DATUM_INT_BIT_1PPS  |\
								  DATUM_INT_BIT_GPS )

/* Time data status bit field definitions */
#define DATUM_STATUS_BIT_FLY	0x01
#define DATUM_STATUS_BIT_TOFS	0x02
#define DATUM_STATUS_BIT_FOFS	0x04

/* Device register acknowledge field bit definitions */
#define DATUM_ACK_BIT_REC		0x01
#define DATUM_ACK_BIT_GPS		0x02
#define DATUM_ACK_BIT_CMD		0x80

/* Dual port ram commands */
#define DATUM_SET_TIMING_MODE_CMD	0x10
#	define DATUM_MODE_TIMECODE		0x00
#	define DATUM_MODE_FREERUN		0x01
#	define DATUM_MODE_EXT1PPS		0x02
#	define DATUM_MODE_RTC			0x03
#	define DATUM_MODE_GPS			0x06

#define DATUM_SET_TIME_FORMAT_CMD	0x11
#	define DATUM_TIME_FMT_BCD		0x00
#	define DATUM_TIME_FMT_UNIX		0x01

#define DATUM_SET_TIME_CMD			0x12

#define DATUM_SET_YEAR_CMD			0x13

#define DATUM_SET_PERIODIC_CMD		0x14
#	define DATUM_NO_SYNC_PPS		0x00
#	define DATUM_SYNC_PPS			0x01

#define DATUM_SET_TC_FORMAT_CMD		0x15
#	define DATUM_TIME_CODE_IRIGA	0x41
#	define DATUM_TIME_CODE_IRIGB	0x42
#	define DATUM_TIME_CODE_1344		0x49

#define DATUM_SET_TC_MOD_CMD		0x16
#	define DATUM_MODE_AMP			0x4D
#	define DATUM_MODE_PULSE			0x44

#define DATUM_SET_TIMING_OFFSET_CMD	0x17

#define DATUM_REQ_UTC_TM_DATA_CMD	0x18
#	define DATUM_GPS_TIME_UTC		0x00
#	define DATUM_GPS_TIME_GPS		0x01
#	define DATUM_LEAP_DELETE		0xFF
#	define DATUM_LEAP_NONE			0x00
#	define DATUM_LEAP_ADD			0x01

#define DATUM_GET_DATA_CMD			0x19
#	define DATUM_GET_DATA_MODE		0x10
#	define DATUM_GET_DATA_TFORMAT	0x11
#	define DATUM_GET_DATA_YEAR		0x13
#	define DATUM_GET_DATA_PER		0x14
#	define DATUM_GET_DATA_TCFMT		0x15
#	define DATUM_GET_DATA_TCMOD		0x16
#	define DATUM_GET_DATA_TOFF		0x17
#	define DATUM_GET_DATA_UTCINFO	0x18

#	define DATUM_GET_DATA_TCOUTFMT	0x1B
#	define DATUM_GET_DATA_TCGENOFF	0x1C
#	define DATUM_GET_DATA_LOCTMOFF	0x1D
#	define DATUM_GET_DATA_LEAPSEC	0x1E

#	define DATUM_GET_DATA_FWVER		0x1F
#	define DATUM_GET_DATA_CLKSRC	0x20
#	define DATUM_GET_DATA_JAMSC		0x21
#	define DATUM_GET_DATA_OSCCTL	0x23
#	define DATUM_GET_DATA_DAVAL		0x24

#	define DATUM_GET_DATA_BATTSTAT	0x26
#	define DATUM_GET_DATA_CLKVAL	0x29
#	define DATUM_GET_DATA_DTFW		0x4F
#	define DATUM_GET_DATA_ASMB		0xF4
#	define DATUM_GET_DATA_HWFAB		0xF5
#	define DATUM_GET_DATA_TFPMODEL	0xF6
#	define DATUM_GET_DATA_SERIAL	0xFE

#define DATUM_SOFT_RESET_CMD		0x1A

#define DATUM_SET_TC_OUT_FMT_CMD	0x1B

#define DATUM_SET_GEN_TM_OFF_CMD	0x1C
#	define DATUM_HALF_NO			0x00
#	define DATUM_HALF_YES			0x01

#define DATUM_SET_LOC_TM_OFF_CMD	0x1D

#define DATUM_SET_LEAP_SEC_CMD		0x1E

#define DATUM_SEL_CLOCK_CMD			0x20
#	define DATUM_SEL_CLOCK_INT		0x49
#	define DATUM_SEL_CLOCK_EXT		0x45

#define DATUM_JAM_SYNC_ENABLE_CMD	0x21
#	define DATUM_JAMSYNC_DIS		0x00
#	define DATUM_JAMSYNC_EN			0x01

#define DATUM_FORCEJAMSYNC_CMD		0x22

#define DATUM_DISC_ENABLE_CMD		0x23
#	define DATUM_DISC_DIS			0x00
#	define DATUM_DISC_EN			0x01

#define DATUM_SET_DAC_CMD			0x24

#define DATUM_SET_DISC_GAIN_CMD		0x25

#define DATUM_REQ_BATT_STAT_CMD		0x26
#	define DATUM_BATT_FAILED		0x00
#	define DATUM_BATT_OK			0x10

#define DATUM_SYNC_RTC_CMD			0x27

#define DATUM_DISC_BATT_RTC_CMD		0x28

#define DATUM_SET_CLK_VAL_CMD		0x29

#define DATUM_SEND_GPS_PKT_CMD		0x30

#define DATUM_REQ_GPS_PKT_CMD		0x31

#define DATUM_MAN_REQ_GPS_PKT_CMD	0x32

#define DATUM_GPS_TIME_FORMAT_CMD	0x33

#define DATUM_SET_GPS_MD_FLG_CMD	0x34
#	define DATUM_DISABLE			0x00
#	define DATUM_ENABLE				0x01

#define DATUM_SET_LOC_TM_FLG_CMD	0x40

#define DATUM_REQ_DAYLT_FLG_CMD		0x41
#	define DATUM_LOCAL_ENABLE		0x08

#define DATUM_SET_YR_INC_FLG_CMD	0x42

#define DATUM_REQ_DTFW_NUM_CMD		0x4F
#	define DATUM_ID_DTFW			"DT6000"

#define DATUM_REQ_ASSEMB_CMD		0xF4
#	define DATUM_OLD_ASM_PART		12043
#	define DATUM_NEW_ASM_PART		12083

#define DATUM_REQ_HW_FAB_CMD		0xF5
#	define DATUM_OLD_FAB_PART		12042
#	define DATUM_NEW_FAB_PART		12083

#define DATUM_REQ_TFP_MODEL_CMD		0xF6
#	define DATUM_ID_IRIG			"BC635PCI"
#	define DATUM_ID_GPS				"BC637PCI"

#define DATUM_REQ_SER_NUM_CMD		0xFE

union binary_event_time
{
	struct
	{
		unsigned 		u_seconds	: 20;
		unsigned		n_seconds	: 4;
		unsigned		status		: 4;
		unsigned		unused		: 4;
		unsigned		seconds		: 32;
	} bits;
	struct
	{
		unsigned long time0;
		unsigned long time1;
	} words;
};

union binary_strobe_time
{
	struct
	{
		unsigned 		u_seconds	: 20;
		unsigned		n_seconds	: 4;
		unsigned		status		: 4;
		unsigned		unused1		: 4;
		unsigned		seconds		: 22;
		unsigned		unused2		: 10;
	} bits;
	struct
	{
		unsigned long minstrb;
		unsigned long majstrb;
	} words;
};

union decimal_event_time
{
	struct
	{
		unsigned 		u_seconds	: 20;
		unsigned		n_seconds	: 4;
		unsigned		status		: 4;
		unsigned		day8		: 1;
		unsigned		unused		: 3;
		unsigned		seconds		: 8;
		unsigned		minutes		: 8;
		unsigned		hours		: 8;
		unsigned		days		: 8;
	} bits;
	struct
	{
		unsigned long time0;
		unsigned long time1;
	} words;
};

union decimal_strobe_time
{
	struct
	{
		unsigned 		u_seconds	: 20;
		unsigned		n_seconds	: 4;
		unsigned		status		: 4;
		unsigned		unused1		: 4;
		unsigned		seconds		: 8;
		unsigned		minutes		: 8;
		unsigned		hours		: 8;
		unsigned		unused2		: 8;
	} bits;
	struct
	{
		unsigned long minstrb;
		unsigned long majstrb;
	} words;
};

struct
{
	unsigned long timereq;
	unsigned long eventreq;
	unsigned long unlock;
	unsigned long r1;
	unsigned long control;
	unsigned long ack;
	unsigned long mask;
	unsigned long intstat;
	unsigned long minstrb;
	unsigned long majstrb;
	unsigned long r2;
	unsigned long r3;
	unsigned long time0;
	unsigned long time1;
	unsigned long event0;
	unsigned long event1;
} *DEV_reg;

struct
{
	char *input;
	char *output;
	char *GPS;
	char *year;
} DP_ram_ofs;

char *DP_ram;

extern int volatile offset;
extern pid_t timer_proxy;
extern int iid;

#pragma off(stack_check)
pid_t far irig_heartbeat( void )
{
	int status = 0;

	if ( DEV_reg->intstat & DATUM_INT_BIT_ALL )
	{
		if ( DEV_reg->intstat & DATUM_INT_BIT_EVENT )
		{
			PARALLEL_BIT_ON( IRQ );
			PARALLEL_BIT_OFF( IRQ );
			offset++;
			status = timer_proxy;
		}
		DEV_reg->intstat = DATUM_INT_BIT_ALL;
	}
	return( status );
}
#pragma on(stack_check)



void write_dpram_data( char *addr, char *data, int len )
{
	for( ; len > 0; len-- )
		*(addr++) = *(data++);
}

double get_irig_time( void )
{
	union binary_event_time etime;
	double result;

	DEV_reg->timereq = 0;
	etime.words.time0 = DEV_reg->time0;
	result = DEV_reg->time1;
	result += ( etime.bits.u_seconds + etime.bits.n_seconds / 10.0 ) / 1000000.0;

	return result;
}

double get_event_time( void )
{
	union binary_event_time etime;
	double result;

	etime.words.time0 = DEV_reg->event0;
	result = DEV_reg->event1;
	DEV_reg->unlock = 0;
	result += ( etime.bits.u_seconds + etime.bits.n_seconds / 10.0 ) / 1000000.0;

	return result;
}

pid_t one_ms_irig_timer_create( void )
{
	unsigned busnum;
	unsigned devfuncnum;
	char irq;
	char pin;
	long address;
	int fd;
	char buffer[10];
	long unix_time;
	char *timestr;
	short year;
	short n1, n2;
	int status;
	pid_t timer_proxy;

	/* Find PCI device */
	if ( ( status = _CA_PCI_Find_Device( DATUM_PCI_DEVICE_ID, DATUM_PCI_VENDOR_ID, 0, &busnum, &devfuncnum ) ) != PCI_SUCCESS )
	{
		PRINT_ERROR( "Can't find Datum IRIG card", status );
		return( ERROR );
	}

	/* Get IRQ for device */
	if ( ( status = _CA_PCI_Read_Config_Byte( busnum, devfuncnum, offsetof(struct _pci_config_regs, Interrupt_Line), 1, &irq ) ) != PCI_SUCCESS )
	{
		PRINT_ERROR( "Can't read interrupt line from Datum IRIG card", status );
		return( ERROR );
	}

	/* Get PIN for device */
	if ( ( status = _CA_PCI_Read_Config_Byte( busnum, devfuncnum, offsetof(struct _pci_config_regs, Interrupt_Pin), 1, &pin ) ) != PCI_SUCCESS )
	{
		PRINT_ERROR( "Can't read interrupt pin from Datum IRIG card", status );
		return( ERROR );
	}

	/* Get the device register memory address */
	if ( ( status = _CA_PCI_Read_Config_DWord( busnum, devfuncnum, offsetof(struct _pci_config_regs, Base_Address_Regs[0]), 1, (char*)&address ) ) != PCI_SUCCESS )
	{
		PRINT_ERROR( "Can't read base address 0 from Datum IRIG card", status );
		return( ERROR );
	}

	/* Map memory address in program space */
	if ( ( fd = shm_open( "Physical", O_RDWR, 0777 ) ) == -1 )
	{
		PRINT_ERROR( "Can't open physical memory for Datum IRIG card", errno );
		return( ERROR );
	}
	if ( ( DEV_reg = mmap( 0, PAGESIZE, PROT_READ | PROT_WRITE, MAP_SHARED, fd, PCI_MEM_ADDR( address ) ) ) == (void *)-1 )
	{
		PRINT_ERROR( "Can't map physical memory for Datum IRIG card", errno );
		return( ERROR );
	}

	/* Get the dual port ram memory address */
	if ( ( status = _CA_PCI_Read_Config_DWord( busnum, devfuncnum, offsetof(struct _pci_config_regs, Base_Address_Regs[1]), 1, (char*)&address ) ) != PCI_SUCCESS )
	{
		PRINT_ERROR( "Can't read base address 1 from Datum IRIG card", status );
		return( ERROR );
	}

	/* Map memory address into program space */
	if ( ( DP_ram = mmap( 0, PAGESIZE, PROT_READ | PROT_WRITE, MAP_SHARED, fd, PCI_MEM_ADDR( address ) ) ) == (void *)-1 )
	{
		PRINT_ERROR( "Can't map physical memory for Datum IRIG card", errno );
		return( ERROR );
	}

	/* Setup pointer to dual port ram buffers */
	DP_ram_ofs.input  = DP_ram + *(DP_ram + DATUM_DPRAM_TOP - 1) + *(DP_ram + DATUM_DPRAM_TOP - 2) * 256;
	DP_ram_ofs.output = DP_ram + *(DP_ram + DATUM_DPRAM_TOP - 3) + *(DP_ram + DATUM_DPRAM_TOP - 4) * 256;
	DP_ram_ofs.GPS    = DP_ram + *(DP_ram + DATUM_DPRAM_TOP - 5) + *(DP_ram + DATUM_DPRAM_TOP - 6) * 256;
	DP_ram_ofs.year   = DP_ram + *(DP_ram + DATUM_DPRAM_TOP - 7) + *(DP_ram + DATUM_DPRAM_TOP - 8) * 256;

	/* Reset card */
	buffer[0] = DATUM_SOFT_RESET_CMD;
	write_dpram_data( DP_ram_ofs.input + 0, &buffer[0], 1 );
	DEV_reg->ack = DATUM_ACK_BIT_CMD;
	while( !( DEV_reg->ack & DATUM_ACK_BIT_REC ) ) delay( 10 );
	sleep( 2 );	

	/* Set mode to RTC - set the time if on input signal */
	buffer[0] = DATUM_SET_TIMING_MODE_CMD;
	buffer[1] = DATUM_MODE_RTC;
	write_dpram_data( DP_ram_ofs.input + 0, &buffer[0], 2 );
	DEV_reg->ack = DATUM_ACK_BIT_CMD;
	while( !( DEV_reg->ack & DATUM_ACK_BIT_REC ) ) delay( 10 );
	sleep( 2 );

	/* Set mode to irig */
	buffer[0] = DATUM_SET_TIMING_MODE_CMD;
	buffer[1] = DATUM_MODE_TIMECODE;
	write_dpram_data( DP_ram_ofs.input + 0, &buffer[0], 2 );
	DEV_reg->ack = DATUM_ACK_BIT_CMD;
	while( !( DEV_reg->ack & DATUM_ACK_BIT_REC ) ) delay( 10 );
	sleep( 2 );

	/* Set year */
	unix_time = time( NULL );
	timestr = asctime( gmtime( (time_t*)&unix_time ) );
	buffer[0] = DATUM_SET_YEAR_CMD;
	year = htons( atoi( timestr + 20 ) );
	write_dpram_data( DP_ram_ofs.input + 0, &buffer[0], 1 );
	write_dpram_data( DP_ram_ofs.input + 1, (char*)&year, 2 );
	DEV_reg->ack = DATUM_ACK_BIT_CMD;
	while( !( DEV_reg->ack & DATUM_ACK_BIT_REC ) ) delay( 10 );
	sleep( 2 );

	/* Set heartbeat */
	buffer[0] = DATUM_SET_PERIODIC_CMD;
	buffer[1] = DATUM_SYNC_PPS;
	n2 = htons( DUTY_CYCLE );
	n1 = htons( FREQUENCY / DUTY_CYCLE );
	write_dpram_data( DP_ram_ofs.input + 0, &buffer[0], 2 );
	write_dpram_data( DP_ram_ofs.input + 2, (char*)&n1, 2 );
	write_dpram_data( DP_ram_ofs.input + 4, (char*)&n2, 2 );
	DEV_reg->ack = DATUM_ACK_BIT_CMD;
	while( !( DEV_reg->ack & DATUM_ACK_BIT_REC ) ) delay( 10 );
	DEV_reg->control = DATUM_CONTROL_BIT_EVENTEN;
	
	/* Setup proxy for timer */
	if ( ( timer_proxy = qnx_proxy_attach( 0, NULL, 0, -1 ) ) == -1 )
	{
		PRINT_ERROR( "Unable to attach proxy", errno );
		return( ERROR );
	}

	DEV_reg->mask = 0;

	if ( ( iid = qnx_hint_attach( irq, irig_heartbeat, FP_SEG( &timer_proxy ) ) ) == ERROR )
	{
		PRINT_ERROR( "Unable to attach interrupt", errno );
		return( ERROR );
	}
	sh_mem_ptr->sync_data.irq = irq;

	DEV_reg->intstat = DATUM_INT_BIT_ALL;
	DEV_reg->mask = DATUM_INT_BIT_EVENT;

	return( timer_proxy );
}

void one_ms_irig_timer_remove( void )
{
	DEV_reg->mask = 0;
	DEV_reg->intstat = DATUM_INT_BIT_ALL;
}
